/*
 * This code is based on an example provided by Richard Stanford, 
 * and originates at the Sun Java tutorial site.

  Copyright (C) 2000 Amanda J. Stent, stent@cs.rochester.edu

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.
 */

import java.awt.*;
import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import java.util.*;
import java.io.*;

public class DynamicTree extends JPanel implements TreeWillExpandListener, TreeExpansionListener {
    // the actual tree
    protected JTree tree;
    // for error beeps
    private Toolkit toolkit = Toolkit.getDefaultToolkit();
    // to fix foobar tree behavior
    HashSet expanded = new HashSet();

    /* Constructor */
    public DynamicTree() {

	// make a treeModel
        Span rootNode = new Span();
        DialogModel treeModel = new DialogModel(rootNode);
        treeModel.addTreeModelListener(new MyTreeModelListener());

	// make a tree
        tree = new JTree(treeModel);
	tree.addTreeWillExpandListener(this);
	tree.addTreeExpansionListener(this);
        tree.setEditable(true);
        tree.getSelectionModel().setSelectionMode
                (TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
	tree.putClientProperty("JTree.lineStyle", "Angled");

	// I don't know why this next line is here -- AS
	ToolTipManager.sharedInstance().registerComponent(tree);
	
	// renderer draws the tree the way I want it
	MyRenderer renderer = new MyRenderer();
	tree.setCellRenderer(renderer);

	// Drawing stuff
        JScrollPane scrollPane = new JScrollPane(tree);
	scrollPane.setPreferredSize(new Dimension(600, 400));
	scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
	scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
        scrollPane.setBorder(
            BorderFactory.createCompoundBorder(
                BorderFactory.createCompoundBorder(
                                BorderFactory.createTitledBorder("Dialog"),
                                BorderFactory.createEmptyBorder(5,5,5,5)),
                scrollPane.getBorder()));
	add(scrollPane);
    }

    /* File I/O */

    // for loading a new dialog
    public void readDialog(File filename) {
	clear();
	addObjects(linestoUtts(readList(filename)));
	reload();
    }
    
    // read a list represented as a series of lines in a file
    protected static Vector readList(File filename) {
	Vector v = new Vector();
	String temp;
	try {
	    BufferedReader reader = new BufferedReader(new FileReader(filename));
	    while ((temp = reader.readLine()) != null) {
		v.addElement(temp);
	    }
	} catch (FileNotFoundException e) { 
	    System.err.println("File not found:  " + filename);
	} catch (IOException e) {
	}
	return v;
    }

    // make utterance-lines into true utterances
    protected static Vector linestoUtts(Vector list) {
	Vector results = new Vector();
	int i = 0;
	String temp = null;
	String all = null;

	do {
	    temp = (String)list.get(i++);

	    // we have a new utterance
	    if (temp.startsWith("utt")) {
		// its the first utterance
		if(all == null) {
		    temp = temp.substring(temp.indexOf(':')+1);
		    all = temp.trim();
		    // it's not the first utterance;
		    // put the previous utterance in the results
		} else {
		    results.addElement(all);
		    temp = temp.substring(temp.indexOf(':')+1);
		    all = temp.trim();
		}
		// it's not a new utterance
	    } else {
		if(all != null) {
		    all = all.concat(" " + temp.trim());
		}
	    }
	} while(i < list.size());
	results.addElement(all);
	return results;
    } 

    // write out the tree
    protected void writeTree(File filename) {
        try {
            BufferedWriter writer = new BufferedWriter(new FileWriter(filename));
            getRoot().writeNode(writer);
            writer.flush();
            writer.close();
        } catch (IOException e) {
        }
    }	

    // read in the tree
    public void readTree(File filename) {

	Span s;
	String temp;
	int newnodes = 0;
	
	clear();

	try {
	    BufferedReader file = new BufferedReader(new FileReader(filename));

	    fillIn(getRoot(), file.readLine());
	    newnodes = getRoot().getID();
	    getRoot().setLabel(file.readLine());
	    
	    while((temp = file.readLine()) != null) {
		s = new Span();
		fillIn(s, temp);
		if(s.getID() > newnodes) newnodes = s.getID();
		s.setLabel(file.readLine());
		addObject(s.getParent(), s);
	    }
	} catch (IOException e) {
	    System.err.println("Error reading nodes");
	}
	getRoot().setNodes(newnodes+1);
	reload();
    }
    
    // fill in the values in a node from the string from the file
    private void fillIn(Span r, String s) {
	StringTokenizer temp = new StringTokenizer(s, " =<>");
	Integer one;
	Boolean two;

	while(temp.hasMoreTokens()) {
	    String current = temp.nextToken();
	    if(current.equals("id")) {
		one = new Integer(temp.nextToken());
		r.setID(one.intValue());
	    } else if (current.equals("number")) {
		one = new Integer(temp.nextToken());		
		r.setNumber(one.intValue());
	    } else if (current.equals("parent")) {
		one = new Integer(temp.nextToken());
		r.setParent(findbyid(one.intValue()));
	    } else if (current.equals("nucleus")) {
		two = new Boolean(temp.nextToken());
		r.setNucleus(two.booleanValue());
	    } else if (current.equals("discounted")) {
		two = new Boolean(temp.nextToken());
		r.setDiscounted(two.booleanValue());
	    }
	}
    }

    /* Tree manipulation */

    // remove the currently selected node
    public void removeCurrentNode() {
	// get the node to be removed
        TreePath currentSelection = tree.getSelectionPath();
        if (currentSelection != null) {
            Span currentNode = (Span)
                         (currentSelection.getLastPathComponent());

	    // get the parent of this node
            Span parent = (Span)(currentNode.getParent());
            if (parent != null) {
		int index = parent.getIndexOfChild(currentNode);

		// remove this node
                getModel().removeNodeFromParent(currentNode);

		// don't lose all its children!
		moveChildrenUp(parent, currentNode, index);

		// if we remove an utterance, make sure we don't end
		// up skipping utterance numbers
		fixNumbers();
		addCollapsed(currentNode);
		reload();
		return;
            }
        }

        // Either there was no selection, or the root was selected.
        toolkit.beep();
    }

    // Remove all nodes except the root
    public void clear() {
        getRoot().removeAllChildren();
        reload();
    }

    /* AS -- it might be nice to be able to move a node around using the mouse */

    public void moveNode(Span node, Span newparent) {
	moveNode(node, newparent, newparent.getChildCount());
    }
    
    public void moveNode(Span node, Span newparent, int location) {
	getModel().removeNodeFromParent(node);
	addObject(newparent, node, location);
    }

    public void moveChildrenUp(Span parent, Span node, int index) {
	int i = 0;
	int kids = getModel().getChildCount(node);

	while (i++ < kids) {
	    Span child = (Span)(getModel().getChild(node, 0));
	    moveNode(child, parent, index+i-1);
	}
    }
    
    public void moveCurrentNode() {
	// find out how many nodes were selected
	int end = tree.getSelectionCount();
	Span [] spans = new Span[end];

	if(end != 2) {
	    toolkit.beep();
	} else {
	    TreePath[] paths = tree.getSelectionPaths();

	    for(int i = 0; i < end; i++) {
		spans[i] = (Span)(paths[i].getLastPathComponent());
	    }
	    if(spans[1].isLeaf()) {
		toolkit.beep();
		return;
	    } else {
		moveNode(spans[0], spans[1]);
	    }

	    fixChildren(spans[1]);
	    reload();
	}
    }

    // make a new node by copying an existing one
    public void duplicateCurrentNode() {
	// get the node to be copied
        TreePath currentSelection = tree.getSelectionPath();
        if (currentSelection != null) {
            Span currentNode = (Span)
                         (currentSelection.getLastPathComponent());
	    
	    // get the parent of this node
            Span parent = (Span)(currentNode.getParent());
            if (parent != null) {

		// make the new node and add it to this parent
		Span newNode = new Span(currentNode.getNumber(), currentNode.getLabel());
		addObject(parent, newNode, parent.getIndexOfChild(currentNode));

		// make sure all the nodes end up in the right place
		fixNumbers();
		reload();
                return;
            }
        } 

        // Either there was no selection, or the root was selected.
        toolkit.beep();
    }

    // used for adding a node
    public void addNode() {
	// find out how many nodes were selected
	int end = tree.getSelectionCount();

	// no nodes were selected
	if(end == 0) {
	    toolkit.beep();
	} else {
	    Span currentNode, newNode;
	    TreePath[] paths = tree.getSelectionPaths();
	    Span[] spans = new Span[end];
	    
	    // collect all the selected nodes
	    for(int i = 0; i < end; i++) {
		spans[i] = (Span)(paths[i].getLastPathComponent());
	    }

	    Span parent = (Span)(spans[0].getParent());
	    
	    for(int i = 1; i < end; i++) {
		if(!((Span)spans[i].getParent()).equals(parent)) {
		    toolkit.beep();
		    return;
		}
	    }

	    // put the new node down under the parent
	    if(parent != null) {
		int index = parent.getIndexOfChild(spans[0]);
		newNode = addObject(parent, new Span(spans[0].getNumber(), " "), index);
	    } else {
		newNode = addObject(new Span(spans[0].getNumber(), " "));
	    }
	    
	    // remove all selected nodes from the parent
	    for(int i = 0; i < end; i++) {
		moveNode(spans[i], newNode);
	    }
	    fixChildren(newNode);
	    addExpanded(newNode);
	    reload();
	}	    
    }

    public Span addObject(Span child) {
        Span parentNode;
        TreePath parentPath = tree.getSelectionPath();

        if (parentPath == null) {
            parentNode = getRoot();
        } else {
            parentNode = (Span)
                         (parentPath.getLastPathComponent());
        }

        return addObject(parentNode, child, parentNode.getChildCount());
    }

    public Span addObject(Span parent, Span child) {
	
	if(parent == null) {
	    parent = getRoot();
	}

        return addObject(parent, child, parent.getChildCount());
    }

    public Span addObject(Span child, int location) {
	return addObject(getRoot(), child, location);
    }

    public Span addObject(Span parent, Span child, int location) {

        if (parent == null) {
            parent = getRoot();
        }

        getModel().insertNodeInto(child, parent, location);

        return child;
    }

    public void addObjects(Vector list) {
	addObjects(null, list);
    }

    public void addObjects(Span parent, Vector list) {
	int i = 0;

	clear();

	do {
	    addObject(parent, new Span(i, (String)list.get(i++)));
	} while (i < list.size());
    }

    public Span findbyid(int id) {
	return findbyidh(getRoot(), id);
    }
    
    public Span findbyidh(Span node, int id) {
	Span temp = null;
	int i = 0;
	
	if(node.getID() == id) {
	    return node;
	} else {
	    while((i < node.getChildCount()) && (temp == null)) {
		temp = findbyidh(node.getChildAt(i++), id);
	    }
	    return temp;
	}
    }

    /* Argumentation-act specific functions */

    // Add a new relation name to a node
    public void addRelation(String name, boolean contentish) {
	TreePath currentSelection = tree.getSelectionPath();
	if (currentSelection != null) {
	    Span currentNode = (Span)
		(currentSelection.getLastPathComponent());
	    if(!currentNode.isLeaf()) {
		currentNode.addText(name, contentish);
		getModel().reload(currentNode);
	    } else {
		// can't add relations to leaves 
		toolkit.beep();
	    }
	} else {
	    // Either there was no selection, or the root was selected.
	    toolkit.beep();
	}
    }

    // make a node a nucleus (or not)
    public void nucleify() {
	// get the selected node
        TreePath currentSelection = tree.getSelectionPath();
        if (currentSelection != null) {
            Span currentNode = (Span)
                         (currentSelection.getLastPathComponent());
	    
	    // (de-)nucleify this node
	    if(currentNode.getNucleus() == false) {
		currentNode.setNucleus(true);
	    } else {
		currentNode.setNucleus(false);
	    }
	    getModel().reload(currentNode.getParent());
	    return;
	}

        // Either there was no selection, or the root was selected.
        toolkit.beep();
    }

    public void discount() {
	// get the selected node
        TreePath currentSelection = tree.getSelectionPath();
        if (currentSelection != null) {
            Span currentNode = (Span)
                         (currentSelection.getLastPathComponent());
	    
	    // (de-)nucleify this node
	    if(currentNode.isLeaf()) {
		if(currentNode.getDiscounted() == false) {
		    currentNode.setDiscounted(true);
		} else {
		    currentNode.setDiscounted(false);
		}
		getModel().reload(currentNode.getParent());
		return;
	    }
	}

        // Either there was no selection, or the root was selected.
        toolkit.beep();
    }

    /* Dialog-specific functions */

    // make sure the utterance numbers stay in order with no gaps
    public void fixNumbers() {
	getRoot().fixNumbers(0);
    }

    public void fixChildren(Span node) {
	int c = node.getChildCount();
	Span kids[] = new Span[c];

	for(int i = 0; i < c; i++) {
	    kids[i] = node.getChildAt(0);
	    getModel().removeNodeFromParent(kids[i]);
	}
	kids = sortSpans(kids, c);
	for(int i = 0; i < c; i++) {
	    addObject(node, kids[i]);
	}
    }
 
    // sort Spans by number using swap-ish sort
    public Span[] sortSpans(Span[] spans, int length) {
	Span temp;

	for(int i = 0; i < length; i++) {
	    int currentlabel = spans[i].getNumber();
	    for(int j = i+1; j < length; j++) {
		if(spans[j].getNumber() < currentlabel) {
		    temp = spans[i];
		    spans[i] = spans[j];
		    spans[j] = temp;
		}
	    }
	}
	return spans;
    }

    /* Various service thingies */

    public DialogModel getModel() { return (DialogModel)tree.getModel(); }
    public Span getRoot() { return (Span)(getModel().getRoot()); }

    class MyTreeModelListener implements TreeModelListener {
        public void treeNodesChanged(TreeModelEvent e) {
            try {
		tree.expandPath(e.getTreePath());
	    } catch (NullPointerException exc) {}
        }
        public void treeNodesInserted(TreeModelEvent e) {
	    tree.expandPath(e.getTreePath());
	}
        public void treeNodesRemoved(TreeModelEvent e) {
        }
        public void treeStructureChanged(TreeModelEvent e) {
        }
    }

    void reload() {
	getModel().reload();
	Iterator it = expanded.iterator();
	while(it.hasNext()) {
	    tree.expandPath(new TreePath(((Span)it.next()).getPath()));
	}
    }
	
    public void addExpanded(Object o) {
	expanded.add(o);
    }

    public void addCollapsed(Object o) {
	expanded.remove(o);
	Span s = (Span)o;

	for(int i = 0; i < s.getChildCount(); i++) {
	    addCollapsed(s.getChildAt(i));
	}
    }

    public HashSet getExpanded() {
	return expanded;
    }

    // draws the nodes
    class MyRenderer extends DefaultTreeCellRenderer {
	ImageIcon tutorialIcon;
	ImageIcon regularIcon;
	ImageIcon discountedIcon;

	public MyRenderer() {
	    tutorialIcon = new ImageIcon("diamond2.gif");
	    regularIcon = new ImageIcon("diamond1.gif");
	    discountedIcon = new ImageIcon("ball_blue.gif");
	}
	
	public Component getTreeCellRendererComponent(
						      JTree tree,
						      Object value,
						      boolean sel,
						      boolean expanded,
						      boolean leaf,
						      int row,
						      boolean hasFocus) {
	    
	    super.getTreeCellRendererComponent(
					       tree, value, sel,
					       expanded, leaf, row,
					       hasFocus);
	    if (isNucleus(value)) {
		setIcon(tutorialIcon);
		setToolTipText(null);
		/*if(!leaf && !(((Span)value).isRoot())) {
		    String temp = makeChildrenString(value);
		    setToolTipText(temp);
		    } */
	    } else if (isDiscounted(value)) {
		setIcon(discountedIcon);
		setToolTipText(null);
	    } else if (!leaf) {
		setIcon(regularIcon);
	    } else {
		setIcon(null);
		setToolTipText(null);
	    }
	    
	    return this;
	}
	
	protected boolean isNucleus(Object value) {
	    return ((Span)value).getNucleus();
	}

	protected boolean isDiscounted(Object value) {
	    return ((Span)value).getDiscounted();
	}

	protected String makeChildrenString(Object value) {
	    String temp = new String("");
	    Span v = (Span)value;
	    
	    int length = v.getChildCount();
	    for(int i = 0; i < length; i++) {
		temp = temp.concat(v.getChildAt(i).toString());
		temp = temp.concat("\n");
	    }
	    return temp;
	}
    }

    public void treeExpanded(TreeExpansionEvent e) {
	//System.out.println("Tree expanded!");
	addExpanded(e.getPath().getLastPathComponent());
    }
    public void treeCollapsed(TreeExpansionEvent e) {
	//System.out.println("Tree collapsed!");
	addCollapsed(e.getPath().getLastPathComponent());
    }
    public void treeWillExpand(TreeExpansionEvent e) {
    }
    public void treeWillCollapse(TreeExpansionEvent e) {    
    }
}



