/*
  Copyright (C) 2000 Amanda J. Stent, stent@cs.rochester.edu

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.
 */

import java.util.*;
import javax.swing.tree.*;
import java.io.*;

public class Span {
    // what is the utterance number (leaf), OR what is the utterance number
    // of the first child (internal node)
    private int number;
    // utterance text (leaf), OR relation labels (internal node)
    private String label;
    // is it a nucleus?
    private boolean nucleus;
    private boolean discounted;
    // each node gets a unique id
    int id;
    // for computing ids
    static int nodes;
    // what is its parent?
    Span parent;
    // what are its children?
    LinkedList children;

    /* Set-up functions */
    private void initialize() {
	number = 0;
	label = " ";
	nucleus = false;
	parent = null;
	children = new LinkedList();
	id = nodes++;
    }

    public Span() {
	initialize();
    }

    public Span(int n) {
	initialize();
	setNumber(n);
    }

    public Span(int n, String l) {
	initialize();
	setNumber(n);
	setLabel(l);
    }

    /* Basic accessor functions */

    public void setNodes(int n) { nodes = n; }

    public void setUserObject(String object) { setLabel(object); }

    public boolean getAllowsChildren() { return true; }

    public void setNumber(int n) { number = n; }

    public int getNumber() { return number; }

    public void setID(int n) { id = n; }

    public int getID() { return id; }

    public void setLabel(String l) { label = l; }
    
    public String getLabel() { return label; }

    public void setNucleus(boolean n) { nucleus = n; }

    public boolean getNucleus() { return nucleus; }

    public void setDiscounted(boolean n) { discounted = n; }

    public boolean getDiscounted() { return discounted; }

    public Span getParent() { return parent; }

    public void setParent(Span p) { parent = p; }
   
    public int getChildCount() { return children.size(); }
   
    public Span getChildAt(int i) { return (Span)children.get(i); }

    public int getIndexOfChild(Span kid) { return children.indexOf(kid); }

    /* File I/O */

    public void writeNode(BufferedWriter w) {
	String temp;
	Integer num = new Integer(0);

	try {
	    w.write('<');
	    w.write("id=", 0, 3);
	    temp = num.toString(id);
	    w.write(temp, 0, temp.length());
	    w.write(' ');
	    
	    w.write("number=", 0, 7);
	    temp = num.toString(number);
	    w.write(temp, 0, temp.length());
	    w.write(' ');

	    if(parent != null) {
		w.write("parent=", 0, 7);
		temp = num.toString(parent.getID());
		w.write(temp, 0, temp.length());
		w.write(' ');
	    }

	    w.write("nucleus=", 0, 8);
	    if(nucleus == true) {
		w.write("true", 0, 4);
	    } else {
		w.write("false", 0, 5);
	    }
	    w.write(' ');

	    w.write("discounted=", 0, 11);
	    if(discounted == true) {
		w.write("true", 0, 4);
	    } else {
		w.write("false", 0, 5);
	    }

	    w.write('>');
	    w.newLine();

	    w.write(label, 0, label.length());
	    w.newLine();

	} catch (IOException e) {
	    System.err.println("Could not finish writing node");
	}

	int i = 0;

	while(i < getChildCount()) {
	    getChildAt(i++).writeNode(w);
	}

    }

    /* Other functions */

    public boolean isRoot() { return (parent == null); }

    public boolean isLeaf() { return (children.size() == 0); }

    public void insertChildAt(Span child, int at) { children.add(at, child); }

    public void remove(int index) {
	// ** Check this -- AS
	children.remove(index);
    }

    public void removeFromParent() {
	parent.removeChild(this);
	parent = null;
    }

    public void removeChild(Span child) { children.remove(child); }

    public void removeAllChildren() { children.clear(); }

    public Span[] getPath() {
	Vector ancestors = new Vector();

	ancestors.add(this);
	Span p = getParent();
	while(p != null) {
	    ancestors.add(p);
	    p = p.getParent();
	}
	int length = ancestors.size();
	Span[] s = new Span[length];
	for(int i = length; i > 0; i--) {
	    //s[i-1] = (Span)ancestors.get(i-1);
	    s[length-i] = (Span)ancestors.get(i-1);
	}
	return s;
    }
    
    public int fixNumbers(int greatestSoFar) {
	int temp = greatestSoFar;
	if(getChildCount() > 0) {
	    for(int i = 0; i < getChildCount(); i++) {
		temp = getChildAt(i).fixNumbers(temp);
	    }
	    setNumber(getChildAt(0).getNumber());
	} else {
		setNumber(temp++);
	}
	return temp;
    }


    public void addText(String l, boolean contentish) {
	int index = label.indexOf(l);
	int index2 = label.indexOf("|", index);
	if(index >= 0) {
	    String before = label.substring(0, index-1);
	    String after = label.substring(index2+1);
	    label = before.concat(after);
	} else if (contentish) {
	     String selectedName = ListDialog.showDialog(null,
							     null);
	     label = label.concat(l + " content: " + selectedName + " | ");
	} else {
	    label = label.concat(l + " | ");
	}
    }

    public String toString() {
	if(isLeaf() && (!label.equals(" "))) {
	    String temp = "utt";
	    temp = temp.concat(number + ": " + label);
	    return temp;
	} else {
	    return label;
	}
    }
}



